#pragma title ("star.c")
/******************************************************************
*
*                           star.c                                *
*                                                                 *
*    Stop Think Act Review self checking panel control program    *
*                                                                 *
*                          Version 2.0                            *
*                                                                 *
*    written by:                                                  *
*                                                                 *
*       United Controls International                             *
*       205 Scientific Drive									  *
*       Norcross, Ga. 30092                                       *
*                                                                 *
******************************************************************/

/* includes */

#include <stdio.h>
#include <conio.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include "ACCES32.h"

/* prototypes */

void read_switches(void);
void write_lamps(void);
/* defines and constants */

#define     ALARM_PORT  3                    /* error alarm */
#define     ALARM_BIT   0x80
//#define     byte        unsigned char
typedef     unsigned char byte;
        
#define     FALSE       0
const byte  flash_map[6][8][2][2] = {              /* flash on setup error */
            0,1, 0,0,               /* I0.0 */     /* see code for explanation */
            0,3, 0,2,               /* I0.1 */ 
            0,5, 0,4,               /* I0.2 */
            1,0, 0,7,               /* I0.3 */
            1,2, 1,1,               /* I0.4 */
            1,4, 1,3,               /* I0.5 */
            1,5, 1,5,               /* I0.6 */
            2,4, 2,3,               /* I0.7 */
            2,6, 2,5,               /* I1.0 */
            3,0, 2,7,               /* I1.1 */
            3,2, 3,2,               /* I1.2 */
            3,1, 3,1,               /* I1.3 */
            3,4, 3,4,               /* I1.4 */
            3,3, 3,3,               /* I1.5 */
            3,6, 3,6,               /* I1.6 */
            3,5, 3,5,               /* I1.7 */
            0,0, 0,0,               /* I2.0 is N/A */
            0,0, 0,0,               /* I2.1 is N/A */
            0,0, 0,0,               /* I2.2 is N/A */
            0,0, 0,0,               /* I2.3 is N/A */
            0,0, 0,0,               /* I2.4 is N/A */
            0,0, 0,0,               /* I2.5 is N/A */
            0,0, 0,0,               /* I2.6 is N/A */
            0,0, 0,0,               /* I2.7 is N/A */
            4,1, 4,1,               /* I3.0 */
            4,0, 4,0,               /* I3.1 */
            4,3, 4,2,               /* I3.2 */
            4,5, 4,4,               /* I3.3 */
            4,7, 4,6,               /* I3.4 */
            5,1, 5,1,               /* I3.5 */
            5,0, 5,0,               /* I3.6 */
            5,3, 5,2,               /* I3.7 */
            5,5, 5,4,               /* I4.0 */
            5,7, 5,7,               /* I4.1 */
            6,0, 6,1,               /* I4.2 */
            6,3, 6,2,               /* I4.3 */
            6,5, 6,4,               /* I4.4 */
            6,7, 6,6,               /* I4.5 */
            0,0, 0,0,               /* I4.6 is N/A */
            7,1, 7,1,               /* I4.7 */
            7,0, 7,0,               /* I5.0 */
            7,3, 7,2,               /* I5.1 */
            7,5, 7,4,               /* I5.2 */
            7,7, 7,6,               /* I5.3 */
            0,0, 0,0,               /* I5.4 is N/A */
            0,0, 0,0,               /* I5.5 is N/A */
            0,0, 0,0,               /* I5.6 is N/A */
            0,0, 0,0};              /* I5.7 is N/A */
const byte  flash_mask[6] = {0xFF, 0xFF, 0, 0xFF, 0xBF, 0x0F};
#define     INBASE      0x290                /* base address for PCDIO48 */
const int   INPORT[6] = {INBASE+0, INBASE+1, INBASE+2,
                        INBASE+4, INBASE+5, INBASE+6};
#define     OT_PORT     0                    /* overtemperature indicator */
#define     OT_BIT      0x40
#define     OUTBASE1    0x280
#define     OUTBASE2    0x288
const int   OUTPORT[8] = {OUTBASE1+0, OUTBASE1+1, OUTBASE1+2, OUTBASE1+3,
                        OUTBASE2+0, OUTBASE2+1, OUTBASE2+2, OUTBASE2+3};
#define     SETUP_PORT  4                    /* setup/run switch */
#define     SETUP_BIT   0x40
#define     TRUE        0xFFFF
   
const char MATCH[] = "CLOSEOPENOFFONSTOPSTARTPRESSRELEASERESET";
#define  CLOSE    0
#define  OPEN     5
#define  OFF      9
#define  ON       12
#define  STOP     14
#define  START    18
#define  PRESS    23
#define  RELEASE  28
#define  RESET    35
#define  INPUT    0
#define  OUTPUT   1

struct SDEF
   {
   char     *label;                          /* option string */
   int      tag;                             /* device position tag */
   int      type;                            /* input or output */
   int      port;
   int      bit;
   int      logic;
   };
const struct SDEF switchdef[] = {
   "S00", CLOSE,   INPUT,  0, 0, 0,
   "S00", OPEN,    INPUT,  0, 0, 1,
   "S01", CLOSE,   INPUT,  0, 1, 0,
   "S01", OPEN,    INPUT,  0, 1, 1,
   "S02", CLOSE,   INPUT,  0, 2, 0,
   "S02", OPEN,    INPUT,  0, 2, 1,
   "S03", CLOSE,   INPUT,  0, 3, 0,
   "S03", OPEN,    INPUT,  0, 3, 1,
   "S04", STOP,    INPUT,  0, 4, 0,
   "S04", START,   INPUT,  0, 4, 1,
   "S05", STOP,    INPUT,  0, 5, 0,
   "S05", START,   INPUT,  0, 5, 1,
   "S06", ON,      INPUT,  0, 6, 0,
   "S06", OFF,     INPUT,  0, 6, 1,
   "S07", CLOSE,   INPUT,  0, 7, 0,
   "S07", OPEN,    INPUT,  0, 7, 1,
   "S10", CLOSE,   INPUT,  1, 0, 0,
   "S10", OPEN,    INPUT,  1, 0, 1,
   "S11", CLOSE,   INPUT,  1, 1, 0,
   "S11", OPEN,    INPUT,  1, 1, 1,
   "S12", CLOSE,   INPUT,  1, 2, 1,
   "S12", OPEN,    INPUT,  1, 2, 0,
   "S12", RELEASE, INPUT,  1, 2, 1,
   "S12", CLOSE,   INPUT,  1, 3, 0,
   "S12", OPEN,    INPUT,  1, 3, 1,
   "S12", RELEASE, INPUT,  1, 3, 1,
   "S14", CLOSE,   INPUT,  1, 4, 1,
   "S14", OPEN,    INPUT,  1, 4, 0,
   "S14", RELEASE, INPUT,  1, 4, 1,
   "S14", CLOSE,   INPUT,  1, 5, 0,
   "S14", OPEN,    INPUT,  1, 5, 1,
   "S14", RELEASE, INPUT,  1, 5, 1,
   "S16", CLOSE,   INPUT,  1, 6, 1,
   "S16", OPEN,    INPUT,  1, 6, 0,
   "S16", RELEASE, INPUT,  1, 6, 1,
   "S16", CLOSE,   INPUT,  1, 7, 0,
   "S16", OPEN,    INPUT,  1, 7, 1,
   "S16", RELEASE, INPUT,  1, 7, 1,
   "S30", CLOSE,   INPUT,  3, 0, 1,
   "S30", OPEN,    INPUT,  3, 0, 0,
   "S30", OFF,     INPUT,  3, 0, 1,
   "S30", CLOSE,   INPUT,  3, 1, 0,
   "S30", OPEN,    INPUT,  3, 1, 1,
   "S30", OFF,     INPUT,  3, 1, 1,
   "S32", CLOSE,   INPUT,  3, 2, 0,
   "S32", OPEN,    INPUT,  3, 2, 1,
   "S33", STOP,    INPUT,  3, 3, 0,
   "S33", START,   INPUT,  3, 3, 1,
   "S34", STOP,    INPUT,  3, 4, 0,
   "S34", START,   INPUT,  3, 4, 1,
   "S35", STOP,    INPUT,  3, 5, 1,
   "S35", START,   INPUT,  3, 5, 0,
   "S35", RELEASE, INPUT,  3, 5, 1,
   "S35", STOP,    INPUT,  3, 6, 0,
   "S35", START,   INPUT,  3, 6, 1,
   "S35", RELEASE, INPUT,  3, 6, 1,
   "S37", CLOSE,   INPUT,  3, 7, 0,
   "S37", OPEN,    INPUT,  3, 7, 1,
   "S40", CLOSE,   INPUT,  4, 0, 0,
   "S40", OPEN,    INPUT,  4, 0, 1,
   "S41", PRESS,   INPUT,  4, 1, 0,
   "S41", START,   INPUT,  4, 1, 0,
   "S41", RELEASE, INPUT,  4, 1, 1,
   "S42", PRESS,   INPUT,  4, 2, 0,
   "S42", RESET,   INPUT,  4, 2, 0,
   "S42", RELEASE, INPUT,  4, 2, 1,
   "S43", STOP,    INPUT,  4, 3, 0,
   "S43", START,   INPUT,  4, 3, 1,
   "S44", CLOSE,   INPUT,  4, 4, 0,
   "S44", OPEN,    INPUT,  4, 4, 1,
   "S45", CLOSE,   INPUT,  4, 5, 0,
   "S45", OPEN,    INPUT,  4, 5, 1,
   "S47", CLOSE,   INPUT,  4, 7, 1,
   "S47", OPEN,    INPUT,  4, 7, 0,
   "S47", OFF,     INPUT,  4, 7, 1,
   "S47", CLOSE,   INPUT,  5, 0, 0,
   "S47", OPEN,    INPUT,  5, 0, 1,
   "S47", OFF,     INPUT,  5, 0, 1,
   "S51", CLOSE,   INPUT,  5, 1, 0,
   "S51", OPEN,    INPUT,  5, 1, 1,
   "S52", CLOSE,   INPUT,  5, 2, 0,
   "S52", OPEN,    INPUT,  5, 2, 1,
   "S53", CLOSE,   INPUT,  5, 3, 0,
   "S53", OPEN,    INPUT,  5, 3, 1,

   "L00", OFF,     OUTPUT, 0, 0, 0,
   "L00", ON,      OUTPUT, 0, 0, 1,
   "L01", OFF,     OUTPUT, 0, 1, 0,
   "L01", ON,      OUTPUT, 0, 1, 1,
   "L02", OFF,     OUTPUT, 0, 2, 0,
   "L02", ON,      OUTPUT, 0, 2, 1,
   "L03", OFF,     OUTPUT, 0, 3, 0,
   "L03", ON,      OUTPUT, 0, 3, 1,
   "L04", OFF,     OUTPUT, 0, 4, 0,
   "L04", ON,      OUTPUT, 0, 4, 1,
   "L05", OFF,     OUTPUT, 0, 5, 0,
   "L05", ON,      OUTPUT, 0, 5, 1,
   "L06", OFF,     OUTPUT, 0, 6, 0,
   "L06", ON,      OUTPUT, 0, 6, 1,
   "L07", OFF,     OUTPUT, 0, 7, 0,
   "L07", ON,      OUTPUT, 0, 7, 1,
   "L10", OFF,     OUTPUT, 1, 0, 0,
   "L10", ON,      OUTPUT, 1, 0, 1,
   "L11", OFF,     OUTPUT, 1, 1, 0,
   "L11", ON,      OUTPUT, 1, 1, 1,
   "L12", OFF,     OUTPUT, 1, 2, 0,
   "L12", ON,      OUTPUT, 1, 2, 1,
   "L13", OFF,     OUTPUT, 1, 3, 0,
   "L13", ON,      OUTPUT, 1, 3, 1,
   "L14", OFF,     OUTPUT, 1, 4, 0,
   "L14", ON,      OUTPUT, 1, 4, 1,
   "L15", OFF,     OUTPUT, 1, 5, 0,
   "L15", ON,      OUTPUT, 1, 5, 1,
   "L16", OFF,     OUTPUT, 1, 6, 0,
   "L16", ON,      OUTPUT, 1, 6, 1,
   "L17", OFF,     OUTPUT, 1, 7, 0,
   "L17", ON,      OUTPUT, 1, 7, 1,
   "L20", OFF,     OUTPUT, 2, 0, 0,
   "L20", ON,      OUTPUT, 2, 0, 1,
   "L21", OFF,     OUTPUT, 2, 1, 0,
   "L21", ON,      OUTPUT, 2, 1, 1,
   "M22", OFF,     OUTPUT, 2, 2, 0,
   "M22", ON,      OUTPUT, 2, 2, 1,
   "L23", OFF,     OUTPUT, 2, 3, 0,
   "L23", ON,      OUTPUT, 2, 3, 1,
   "L24", OFF,     OUTPUT, 2, 4, 0,
   "L24", ON,      OUTPUT, 2, 4, 1,
   "L25", OFF,     OUTPUT, 2, 5, 0,
   "L25", ON,      OUTPUT, 2, 5, 1,
   "L26", OFF,     OUTPUT, 2, 6, 0,
   "L26", ON,      OUTPUT, 2, 6, 1,
   "L27", OFF,     OUTPUT, 2, 7, 0,
   "L27", ON,      OUTPUT, 2, 7, 1,
   "L30", OFF,     OUTPUT, 3, 0, 0,
   "L30", ON,      OUTPUT, 3, 0, 1,
   "L31", OFF,     OUTPUT, 3, 1, 0,
   "L31", ON,      OUTPUT, 3, 1, 1,
   "L32", OFF,     OUTPUT, 3, 2, 0,
   "L32", ON,      OUTPUT, 3, 2, 1,
   "L33", OFF,     OUTPUT, 3, 3, 0,
   "L33", ON,      OUTPUT, 3, 3, 1,
   "L34", OFF,     OUTPUT, 3, 4, 0,
   "L34", ON,      OUTPUT, 3, 4, 1,
   "L35", OFF,     OUTPUT, 3, 5, 0,
   "L35", ON,      OUTPUT, 3, 5, 1,
   "L36", OFF,     OUTPUT, 3, 6, 0,
   "L36", ON,      OUTPUT, 3, 6, 1,
   "M37", OFF,     OUTPUT, 3, 7, 0,
   "M37", ON,      OUTPUT, 3, 7, 1,
   "L40", OFF,     OUTPUT, 4, 0, 0,
   "L40", ON,      OUTPUT, 4, 0, 1,
   "L41", OFF,     OUTPUT, 4, 1, 0,
   "L41", ON,      OUTPUT, 4, 1, 1,
   "L42", OFF,     OUTPUT, 4, 2, 0,
   "L42", ON,      OUTPUT, 4, 2, 1,
   "L43", OFF,     OUTPUT, 4, 3, 0,
   "L43", ON,      OUTPUT, 4, 3, 1,
   "L44", OFF,     OUTPUT, 4, 4, 0,
   "L44", ON,      OUTPUT, 4, 4, 1,
   "L45", OFF,     OUTPUT, 4, 5, 0,
   "L45", ON,      OUTPUT, 4, 5, 1,
   "L46", OFF,     OUTPUT, 4, 6, 0,
   "L46", ON,      OUTPUT, 4, 6, 1,
   "L47", OFF,     OUTPUT, 4, 7, 0,
   "L47", ON,      OUTPUT, 4, 7, 1,
   "L50", OFF,     OUTPUT, 5, 0, 0,
   "L50", ON,      OUTPUT, 5, 0, 1,
   "L51", OFF,     OUTPUT, 5, 1, 0,
   "L51", ON,      OUTPUT, 5, 1, 1,
   "L52", OFF,     OUTPUT, 5, 2, 0,
   "L52", ON,      OUTPUT, 5, 2, 1,
   "L53", OFF,     OUTPUT, 5, 3, 0,
   "L53", ON,      OUTPUT, 5, 3, 1,
   "L54", OFF,     OUTPUT, 5, 4, 0,
   "L54", ON,      OUTPUT, 5, 4, 1,
   "L55", OFF,     OUTPUT, 5, 5, 0,
   "L55", ON,      OUTPUT, 5, 5, 1,
   "M56", OFF,     OUTPUT, 5, 6, 0,
   "M56", ON,      OUTPUT, 5, 6, 1,
   "L57", OFF,     OUTPUT, 5, 7, 0,
   "L57", ON,      OUTPUT, 5, 7, 1,
   "L60", OFF,     OUTPUT, 6, 0, 0,
   "L60", ON,      OUTPUT, 6, 0, 1,
   "L61", OFF,     OUTPUT, 6, 1, 0,
   "L61", ON,      OUTPUT, 6, 1, 1,
   "L62", OFF,     OUTPUT, 6, 2, 0,
   "L62", ON,      OUTPUT, 6, 2, 1,
   "L63", OFF,     OUTPUT, 6, 3, 0,
   "L63", ON,      OUTPUT, 6, 3, 1,
   "L64", OFF,     OUTPUT, 6, 4, 0,
   "L64", ON,      OUTPUT, 6, 4, 1,
   "L65", OFF,     OUTPUT, 6, 5, 0,
   "L65", ON,      OUTPUT, 6, 5, 1,
   "L66", OFF,     OUTPUT, 6, 6, 0,
   "L66", ON,      OUTPUT, 6, 6, 1,
   "L67", OFF,     OUTPUT, 6, 7, 0,
   "L67", ON,      OUTPUT, 6, 7, 1,
   "L70", OFF,     OUTPUT, 7, 0, 0,
   "L70", ON,      OUTPUT, 7, 0, 1,
   "L71", OFF,     OUTPUT, 7, 1, 0,
   "L71", ON,      OUTPUT, 7, 1, 1,
   "L72", OFF,     OUTPUT, 7, 2, 0,
   "L72", ON,      OUTPUT, 7, 2, 1,
   "L73", OFF,     OUTPUT, 7, 3, 0,
   "L73", ON,      OUTPUT, 7, 3, 1,
   "L74", OFF,     OUTPUT, 7, 4, 0,
   "L74", ON,      OUTPUT, 7, 4, 1,
   "L75", OFF,     OUTPUT, 7, 5, 0,
   "L75", ON,      OUTPUT, 7, 5, 1,
   "L76", OFF,     OUTPUT, 7, 6, 0,
   "L76", ON,      OUTPUT, 7, 6, 1,
   "L77", OFF,     OUTPUT, 7, 7, 0,
   "L77", ON,      OUTPUT, 7, 7, 1
   };
#define NDEFS sizeof(switchdef)/sizeof(switchdef[0])

/* global storage */

int      alarm;                              /* set if error in state */
clock_t  alarm_timeout;                      /* limit how long the alarm sounds */
clock_t  delay;                              /* state timeout delay */
int      delaystate[100];                    /* delay (ticks) - if not 0, go to next state after specified time */
int      done_state;                         /* when this state is reached, we are done */
byte     flash[8];                           /* output array for flashing */
int      flash_on;                           /* status flashing is enabled */
clock_t  flash_start;                        /* flash timer */
byte     incurrent[6];                       /* inputs at current state */
byte     instate[100][6];                    /* switch input states */
byte     outcurrent[8];
byte     outmask[6];                         /* mask active bits for output state */
byte     outstate[100][8];                   /* device output states */
int      overtemp;                           /* set after timeout */
double   overtemp_limit;                     /* time unit overtemp (min) */
int      setup;                              /* status of RUN/SETUP switch */
int      state;                              /* current state */
clock_t  test_start_time;                    /* time test started */
FILE     *infile;                            /* ini file */

void main (int argc, char *argv[])
   {
   int   port;
   int   bit;
   int   actual;
   int   want;
   int   i, j;
   char  *filename;
   char  buf[133];
   char  *ep;
   int   found;
   int   COFlag;
   int   itemp;
   int   tag;
   char  *tagptr;
            
   printf(
   "************************************************\n"
   "*                                              *\n"
   "*       =====Stop Think Act Review====         *\n"
   "*              Procedure Trainer               *\n"
   "*                                              *\n"
   "*                 Version 2.0                  *\n"
   "*                                              *\n"
   "* built by:  United Controls International     *\n"
   "*            (770) 496-1406                    *\n"
   "*                                              *\n"
   "************************************************\n");

   /****** open inititialization file and set parameters *****/
   
   if (argc == 1)                            /* get name of initialization file */
      filename = "default.ini";
   else
      filename = argv[1];
      
   if ((infile = fopen(filename, "rt")) == NULL)
      {
      printf("\a\nUnable to open procedure file %s\n", filename);
      exit(1);
      }
      
   for (i=0; i<100; i++)                     /* clear state arrays */
      {
      for (j=0; j<6; j++)
         instate[i][j] = 0;
      for (j=0; j<8; j++)
         outstate[i][j] = 0;
      delaystate[i] = 0;
      }
         
   while (1)
      {
      if (fgets(buf, 132, infile) == NULL)   /* read line from file */
         if (feof(infile))                   /* done */
            break;
         else
            {
            printf("\a\nError reading procedure file %s\n", filename);
            exit(1);
            }
         
      if (buf[0] == ';')
         continue;                           /* comment */
         
      strupr(buf);                           /* make it upper case */
      buf[strlen(buf)-1] = '\0';             /* get rid of newline */
      if (buf[0] == '[')                     /* this is a section name */
         {
         if (strncmp(buf, "[CABINETOVERTEMPERATURE]", 24) == 0)   /* cabinet overtemperature */
            {
            COFlag = TRUE;
            continue;
            }
         else
            COFlag = FALSE;
            
         i = sscanf(buf+1,"%i", &itemp);    /* assume this is a state # */
         if (i == 0)
            {
            printf("\nUnrecognized section name: %s\n", buf);
            exit(1);
            }
         else
            {
            if ((itemp < 0) || (itemp > 99))
               {
               printf("\nInvalid state number: %s\n", buf);
               exit(1);
               }
            else
               {
               if ((state > 0) && ((state+1) != itemp))
                  {
                  printf("\a\nError at state %i: states must be specified "
                     "in sequence in .INI file\n", itemp);
                  exit(1);
                  }
               if (itemp > 0)                /* copy state to next state */
                  {
                  memcpy(instate[itemp], instate[state], 6);
                  memcpy(outstate[itemp], outstate[state], 8);
                  }
               state = itemp;
               }
            }
            
         continue;                           /* next input line */
         }
         
      if ((ep = strrchr(buf, '=')) == NULL)  /* parse the equal sign */
         continue;                           /* not an entry line */
      ++ep;                                  /* point to option value */

      if (strncmp(buf, "DELAY=", 6) == 0)    /* delay */
         if (!COFlag)
            {
            i = sscanf(buf+6,"%i", &delaystate[state]);
            if (i == 0)
               {
               printf("\nUnrecognized opion: %s\n", buf);
               exit(1);
               }
            continue;                        /* next line */
            }
         else
            {
            i = sscanf(ep,"%i", &itemp);
            if (i == 0)
               {
               printf("\nUnrecognized opion: %s\n", buf);
               exit(1);
               }
            overtemp_limit = (double)itemp / 60.0;
            continue;                        /* next line */
            }

      found = FALSE;
      for (i=0; i<NDEFS; i++)                /* search the definition table */
         {
         if (strncmp(buf, switchdef[i].label, ep-buf-1) == 0)
            {
            if ((tagptr = strstr(MATCH, ep)) == NULL)
               continue;
               
            tag = tagptr - MATCH;
         
            if (tag != switchdef[i].tag)
               continue;
               
            if (switchdef[i].type == INPUT)
               {
               instate[state][switchdef[i].port] &= 
                  (0xFF - (1 << switchdef[i].bit));
               instate[state][switchdef[i].port] |= 
                  (switchdef[i].logic << switchdef[i].bit);
               found = TRUE;
               }

            if (switchdef[i].type == OUTPUT)
               {
               outstate[state][switchdef[i].port] &= 
                  (0xFF - (1 << switchdef[i].bit));
               outstate[state][switchdef[i].port] |= 
                  (switchdef[i].logic << switchdef[i].bit);
               found = TRUE;
               }
            }
         }
      if (!found)
         {
         printf("\n\aUnrecognized option %s in state %i\n", buf, state);
         exit(1);
         }
      }
   fclose(infile);
   done_state = state;

   /***** initialize PCDIO48 card *****/
   
   OutPortB(INBASE+3, 0x9B);                     /* mode 0, A,B,C are inputs */
   OutPortB(INBASE+7, 0x9B);                     /* mode 0, A,B,C are inputs */
   
   
   /***** require RUN/SETUP switch to be in SETUP mode *****/
   
   read_switches();
   if (!setup)
      {
      printf("\n\aPlace RUN/SETUP switch in SETUP position\n");
      
      while (!setup)
         read_switches();
      }

   /***** main setup/run loop *****/       

   while (TRUE)
      {

      /***** setup panel ******/
      
      /* in setup mode, if a switch is not in the correct state 0 position,
         a light associated with that switch will flash */
      
      printf("\nSETUP Mode: if lamp is flashing, change switch position."
             "\nSelect RUN when all lamps are off.");
      flash_on = TRUE;                       /* lights initially on */
      flash_start = clock();
      while (setup)
         {
         read_switches();
         if ((double)(clock() - flash_start)/CLOCKS_PER_SEC > 0.5) /* 1 sec lamp flasher */
            {
            flash_on = !flash_on;
            flash_start = clock();
            }
         else
            continue;
            
         for (i=0; i< 8; i++)                /* reset all bits */
            flash[i] = 0;
            
         /* description of flash_map                                                   */
         /*   flash_map[0]            switch port                                      */
         /*             | [0]           switch bit                                     */
         /*             |  | [0]           if instate[0] bit is 0                      */
         /*             |  |  | [0] = 0       lamp port to flash if switch status is 1 */
         /*             |  |  | [1] = 1       lamp bit to flash if switch status is 1  */
         /*             |  | [1]           if instate[0] bit is 1                      */
         /*             |  |  | [0] = 0       lamp port to flash if switch status is 0 */
         /*             |  |  | [1] = 0       lamp bit to flash if switch status is 0  */
         /*             | [1]                                                          */
         /*             |  :                                                           */
         /*             | [7]                                                          */
         /*            [1]                                                             */
         /*             :                                                              */
         /*            [5]                                                             */

         if (flash_on)                       /* build flash output */
            for (port=0; port<6; port++)
               {
               for (bit=0; bit<8; bit++)
                  {
                  if (!((flash_mask[port] >> bit) & 1))
                     continue;               /* don't flash this bit if masked */
                  for (i=0; i<2; i++)
                     {
                     actual = (incurrent[port] >> bit) & 1;
                     want = (instate[0][port] >> bit) & 1;
                     if (want != i)
                        continue;
                     if (actual != want)
                        flash[flash_map[port][bit][i][0]] 
                           |= (1 << flash_map[port][bit][i][1]);
                     }
                  }
               }
         state = -1;
         write_lamps();
         }

      /*********** begin run ***********/
   
      state = 0;
      alarm = FALSE;
      overtemp = FALSE;
      write_lamps();
   
      printf("\nBeginning test sequence"
             "\nState = 0\n");
   
      /********** main state loop ***********/
   
      while(!alarm && (state < done_state))
         {
         write_lamps();
         
         if (state == 0)
            test_start_time = clock();       /* clock doesn't start til state 1 */
            
         if (!overtemp && (((double)(clock() - test_start_time)/CLOCKS_PER_SEC) > (overtemp_limit * 60.0)))
            {
            overtemp = TRUE;
            printf("\a\a\a15 minute overtemp timeout\n");
            }

         if (delaystate[state] && (clock() > delay))   /* autostep to next state */
            {
            ++state;
            printf("State = %i, Elapsed time = %.1f min\n",
               state, (double)(clock() - test_start_time)/(CLOCKS_PER_SEC*60.0));
            if (delaystate[state])
               delay = clock() + delaystate[state]*CLOCKS_PER_SEC;
            }
                     
         read_switches();
         if (setup)
            break;
      
         for (i=0; i<6; i++)                 /* are we still in current state? */
            if (incurrent[i] != instate[state][i])
               alarm = TRUE;                 /* no - assume alarm */
            
         if (!alarm)
            continue;                        /* yes - go to top of loop */

         alarm = 0;
         for (i=0; i<6; i++)                 /* do we match next state? */
            if (incurrent[i] != instate[state+1][i])
               alarm = TRUE;                 /* no - set alarm */
            
         if (!alarm)
            {
            ++state;                         /* yes - update state */
            printf("State = %i, Elapsed time = %.1f min\n",
               state, (double)(clock() - test_start_time)/(CLOCKS_PER_SEC*60.0));
            if (state == 1)
               printf("Beginning %.0f minute overtemp timer\n", overtemp_limit);
            if (delaystate[state])
               delay = clock() + delaystate[state]*CLOCKS_PER_SEC;
            }
         else
            printf("Error in positioning switch\n");
         }
      
      write_lamps();
      
      if (state == done_state)
         printf("\nCongratulations! You followed the procedure correctly!\n"
                "Elapsed time was %.1f minutes\n", 
                (double)(clock() - test_start_time)/(CLOCKS_PER_SEC*60.0));
            
      if (!setup)
         printf("\nPlace RUN/SETUP switch to SETUP to prepare for next run\n");
   
      alarm_timeout = clock();
      
      while(TRUE)                               /* wait until switch is in SETUP */
         {
         if (((double)(clock() - alarm_timeout)/CLOCKS_PER_SEC) > 5)
            alarm = FALSE;                      /* timeout the alarm */
         write_lamps();
         read_switches();
         if (setup)
            break;
         }
      }         
   }
         
      
/************ read_switches() *************/

void read_switches(void)
   {
   int      i;
   int      tick_cnt=0;
   clock_t  last=clock();
   
   while(tick_cnt < 2)                       /* debounce input over at least 2 ticks (114 ms) */
      {
      if (clock() != last)                   /* look for tick */
         {
         ++tick_cnt;                         /* and bump tick counter */
         last = clock();
         }
      for (i=0; i<6; i++)
         {  
         if (incurrent[i] != (byte)InPortB(INPORT[i]))       /* if input has changed */
            {
            incurrent[i] = (byte)InPortB(INPORT[i]);        /* then start over */
            tick_cnt = 0;
            }
         }
      }
   setup = incurrent[SETUP_PORT] & SETUP_BIT; /* extract the setup status */
   for (i=0; i<6; i++)
      incurrent[i] &= flash_mask[i];         /* remove unused inputs */
      
   if (kbhit())                              /* check for ctrl c exit */
      {
      if (getch() == 3)
         {
         for (i=0; i<8; i++)                 /* turn off lamps */
            OutPortB(OUTPORT[i], 0);
         exit(0);
         }
      }
   }
   
/************* write_lamps() ***************/

void write_lamps()
   {
   int   i;
   
   if (state == -1)                          /* display flash output */
      {
      for (i=0; i<8; i++)
         OutPortB(OUTPORT[i], flash[i]);
      }
   else                                      /* display state output */
      {
      for (i=0; i<8; i++)
         {
         outcurrent[i] = outstate[state][i];
         if (i == OT_PORT)                    /* include overtemp light */
            {
            if (overtemp)
               outcurrent[i] |= OT_BIT;
            else
               outcurrent[i] &= ~OT_BIT;
            }
         if (i == ALARM_PORT)                 /* include error alarm */
            {
            if (alarm)
               outcurrent[i] |= ALARM_BIT;
            else
               outcurrent[i] &= ~ALARM_BIT;
            }
         OutPortB(OUTPORT[i], outcurrent[i]);
         }
      }
   }
   